; File:		updi-tiny3216\main.asm
; Device:	ATtiny3216
; Created:	09.06.2024 21:52:27
; Version:	2024-06-25
; Author:	Johannes Fechner

.nolist
.include <tn3216def.inc>
.include "macros.asm"
.list

; == Hardware definitions ==
.nolist
.include "hw-sldt.asm"
.list

; == PC UART definitions ==
.equ UART_BPS = 115_200 ; bits/s

; == UPDI protocol definitions ==
.equ UPDI_BPS = 50_000		; UPDI UART baud rate in bits/s.
.equ UPDI_UART_BUFLEN = 16	; UPDI UART tx/rx buffer length in bytes.

; == Register name definitions ==
.def zero = r2		; Contains always 0x00.

; === Temporary registers ===
.def tmp0 = r16
.def tmp1 = r17
.def tmp2 = r18
.def tmp3 = r19
.def tmp4 = r20
.def tmpL = r24
.def tmpH = r25

; === Main registers ===
.def updiFlags = r3				; UPDI status bits and flags.
.equ UPDI_FLAGS_BUSY_bp = 0		; UPDI transmission or reception in progress.
.equ UPDI_FLAGS_TXC_bp = 1		; UPDI transmission complete.
.equ UPDI_FLAGS_RXC_bp = 2		; UPDI Frame Received flag.
.equ UPDI_FLAGS_TIMEOUT_bp = 3	; Timeout occurred while waiting for reception.
.equ UPDI_FLAGS_BUFOVF_bp = 4	; Buffer overflow while receiving.

; == RAM usage ==
.dseg
.org SRAM_START

updiUartBitCntr_ram:		.byte   1
; == 0: during start bit.
; == 1: during data LSb.
; ...
; == 8: during data MSb.
; == 9: during (even) parity bit.
; ==10: during 1st stop bit.
; ==11: during 2nd stop bit.
updiUartBitMask_ram:		.byte   1
updiUartDataBuf_ram:		.byte   UPDI_UART_BUFLEN
updiUartDataCount_ram:		.byte	1
updiUartDataIndex_ram:		.byte	1
updiUartParityCntr_ram:		.byte   1
updiUartTimeoutCntr_ram:	.byte   1

writeAddrL_ram:				.byte   1
writeAddrH_ram:				.byte   1
writeData_ram:				.byte   1

.cseg
.org 0
	rjmp	reset

; == Interrupt vectors ==
.org TARGET_UPDI_PORT_VECT
	rjmp	updi_fallingEdgeIsr
.org TCA0_OVF_vect
	rjmp	updi_timerIsr
	
; == Zero-terminated strings ==
.org INT_VECTORS_SIZE
szMenu: .db "(1) Read SIB", '\r', '\n',\
	"(2) Read UPDI register", '\r', '\n',\
	"(3) Read byte from UPDI address space", '\r', '\n',\
	"(4) Chip erase", '\r', '\n',\
	"(5) Write EEPROM byte", '\r', '\n', 0, 0
szAddress: .db "Address: ", 0
szData: .db " Data: ", 0
szLdcs: .db "LDCS returned ", 0, 0
szLds: .db "LDS returned ", 0
szConfirm: .db " Confirm (y/n): ", 0, 0
szEnteringCEK: .db "Entering Chip Erase key ... ", '\r', '\n', 0, 0
szEnteringNVMPK: .db "Entering NVM Programming key ... ", '\r', '\n', 0
szAsiKeyStatus: .db "ASI KEY STATUS: ", 0, 0
szReset: .db "Issuing System Reset ...", '\r', '\n', 0, 0
szAsiSysStatus: .db "ASI SYS STATUS: ", 0, 0
szNvmctrlStatus: .db "NVMCTRL_STATUS: ", 0, 0
szNvmVersion: .db "NVM version number: ", 0, 0
szSendingSts: .db "Sending STS ... ", '\r', '\n', 0, 0
szOK: .db "OK", '\r', '\n', 0, 0
szWaitNvm: .db "Waiting while NVMCTRL is busy ...", '\r', '\n', 0
szErrInvAnswer: .db "Error: Invalid answer.", '\r', '\n', '\r', '\n', 0, 0
szErrBufOvf: .db "Error: Buffer overflow while receiving.", '\r', '\n', '\r', '\n', 0
szErrTimeout: .db "Error: Timeout while waiting for UPDI answer.", '\r', '\n', '\r', '\n', 0
szErrKey: .db "Error: ASI Key submitting failed.", '\r', '\n', '\r', '\n', 0
szErrErasefail: .db"Error: ERASEFAIL==1 - Chip Erase failed.", '\r', '\n', '\r', '\n', 0, 0
szLineBreak: .db '\r', '\n', 0, 0

; == External routines ==
.nolist
.include "uart.asm"
.include "wait.asm"
.include "updi-phy.asm"
.include "updi-link.asm"
.list

; == Execution entry point after reset ==
reset:
; == Set clock ==
	setClock

; == Initialize registers ==
	clr		zero
	clr		updiFlags

; == Initialize UART ==
.equ BAUD_REG_VAL = ROUND(4.0*F_CPU/(1.0*UART_BPS))
	ldi		tmp0, low(BAUD_REG_VAL)
	xout	USART0_BAUDL, tmp0
	ldi		tmp0, high(BAUD_REG_VAL)
	xout	USART0_BAUDH, tmp0
; USART0_CTRLC = $03 (reset value): asynchronous, 8N1.
	ldi		tmp0, (1<<USART_TXEN_bp) | (1<<USART_RXEN_bp)
	xout	USART0_CTRLB, tmp0

; == Initialize TCA0 for UPDI Soft-UART ==
; Enable overflow interrupt:
	ldi		tmp0, TCA_SINGLE_OVF_bm
	xout	TCA0_SINGLE_INTCTRL, tmp0

; == Initialize I/O ==
; Configure TxD pin as output:
	sbi		TXD_VDIR, TXD_bp
; Configure Rx/Tx LEDs as outputs:
	sbi		LED_UPDI_RX_VDIR, LED_UPDI_RX_bp
	sbi		LED_UPDI_TX_VDIR, LED_UPDI_TX_bp
; Enable pull-up at Target UPDI pin:
	ldi		tmp0, PORT_PULLUPEN_bm
	xout	TARGET_UPDI_PINCTRL, tmp0

; == Enable interrupts ==
	sei
	
loop:
; Print menu to PC UART:
	ldiz	2*szMenu
	rcall	uart_putSz
; Wait for PC UART reception:
	rcall	uart_getChar
	cpi		tmp0, '1'
	brne	loop_2
	rjmp	loop_readSib
loop_2:
	cpi		tmp0, '2'
	brne	loop_3
	rjmp	loop_readUpdiReg
loop_3:
	cpi		tmp0, '3'
	brne	loop_4
	rjmp	loop_readByte
loop_4:
	cpi		tmp0, '4'
	brne	loop_5
	rjmp	loop_chipErase
loop_5:
	cpi		tmp0, '5'
	brne	loop_default
	rjmp	loop_writeEeprom
loop_default:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop

loop_readSib:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_readSib_waitTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_readSib_waitTx
; Wait until all SIB bytes were received:
loop_readSib_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_readSib_waitRx
; updiUartDataBuf_ram now contains the received bytes.
	clr		tmp1
    ldiz	updiUartDataBuf_ram
loop_readSib_print:
	ld		tmp0, Z+
    rcall   uart_putChar
	inc		tmp1
	cpi		tmp1, 16
	brlo	loop_readSib_print
	ldiz	2*szLineBreak
	rcall	uart_putSz
    ldiz    2*szLineBreak
    rcall   uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readUpdiReg:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send LDCS instruction:
	andi	tmp0, 0x0F
	rcall	updi_ldcs
; Wait for answer bytes from Target UPDI:
loop_readUpdiReg_wait:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readUpdiReg_wait
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
	ldiz	2*szLdcs
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop

loop_readByte:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Get address from user:
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte	; Address MSB.
	mov		YH, tmp0
	rcall	uart_getHexByte	; Address LSB.
	mov		YL, tmp0
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_readByte_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_readByte_waitSibTx
; Wait until all SIB bytes were received:
loop_readByte_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_readByte_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Give the target MCU a bit of time:
	rcall	wait25ms
; Branch according to NVM version:
	cpi		tmp0, '0'
	breq	loop_readByte_nvm0
; NVM version number != 0, e.g. AVR_DB and AVR_DD.
; These devices apparently need a more complicated procedure:
	rcall	updi_applyReset
	rcall	updi_idle
	rcall	wait25ms
	rcall	updi_releaseReset
	rcall	updi_idle
	rcall	wait25ms
	rcall	updi_enterNvmProgKey
loop_readByte_nvm0:
	rcall	updi_applyReset
	rcall	updi_idle
	rcall	wait25ms
	rcall	updi_releaseReset
	rcall	updi_idle
	rcall	wait25ms
; Send LDS instruction:
	rcall	updi_lds
; Read answer byte from Target UPDI:
loop_readByte_waitLds:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_readByte_waitLds
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
	ldiz	2*szLds
	rcall	uart_putSz
    rcall   uart_putHexByte
    ldiz    2*szLineBreak
    rcall   uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rcall	updi_disable
	rjmp	loop
	
loop_chipErase:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_chipErase_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_chipErase_waitSibTx
; Wait until all SIB bytes were received:
loop_chipErase_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_chipErase_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	mov		tmp1, tmp0
; Enter the Chip Erase key:
	ldiz	2*szEnteringCEK
	rcall	uart_putSz
	rcall	updi_enterChipEraseKey
	cpi		tmp1, '0'
	breq	loop_chipErase_confirmKey
; Enter the NVMPROG key, too (recommended in AVR64DD28 datasheet):
	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
	rcall	updi_enterNvmProgKey
loop_chipErase_confirmKey:
; == Confirm success by reading ASI Key Status Register ==
	ldi		tmp0, UPDI_ASI_KEY_STATUS
	rcall	updi_ldcs
; Read answer from target UPDI:
loop_chipErase_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitKeyStatus
	lds		tmp0, updiUartDataBuf_ram
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	andi	tmp0, (1<<UPDI_CHIPERASE_bp)
	brne	loop_chipErase_keySuccess
; CHIPERASE key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rjmp	loop_exit
loop_chipErase_keySuccess:
	ldiz	2*szOK
	rcall	uart_putSz
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	updi_applyReset
	rcall	updi_idle
	rcall	wait25ms
	rcall	updi_releaseReset
	rcall	updi_idle
loop_chipErase_readLockStatus:
; Read LOCKSTATUS:
	rcall	wait25ms
	ldi		tmp0, UPDI_ASI_SYS_STATUS
	rcall	updi_ldcs
; Read answer from target UPDI:
loop_chipErase_waitLockStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_chipErase_waitLockStatus
	lds		tmp0, updiUartDataBuf_ram
	sbrc	tmp0, UPDI_LOCKSTATUS_bp
	rjmp	loop_chipErase_readLockStatus
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szLineBreak
	rcall	uart_putSz
	cpi		tmp1, '0'
	breq	loop_chipErase_end
; NVM version '2' provides an ERASEFAIL bit in ASI_SYS_STATUS.
	sbrs	tmp0, UPDI_ERASEFAIL_bp
	rjmp	loop_exit
	ldiz	2*szErrErasefail
	rcall	uart_putSz
loop_chipErase_end:
	rjmp	loop_exit

loop_writeEeprom:
; Echo the entered character:
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
; === Get address from user ===
	ldiz	2*szAddress
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeAddrH_ram, tmp0
	rcall	uart_getHexByte
	sts		writeAddrL_ram, tmp0
; === Get data from user ===
	ldiz	2*szData
	rcall	uart_putSz
	rcall	uart_getHexByte
	sts		writeData_ram, tmp0
	ldiz	2*szConfirm
	rcall	uart_putSz
	rcall	uart_getChar
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	cpi		tmp0, 'y'
	breq	loop_writeEeprom_confirmed
; Operation cancelled by user.
	rjmp	loop
loop_writeEeprom_confirmed:
; Enable Target UPDI:
	rcall	updi_enable
; Send Key Get SIB command, to keep UPDI alive and get the NVM version:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, UPDI_KEY | UPDI_KEY_GET_SIB | UPDI_KEY_128BITS
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for transmission to finish:
loop_writeEeprom_waitSibTx:
	sbrs	updiFlags, UPDI_FLAGS_TXC_bp
	rjmp	loop_writeEeprom_waitSibTx
; Wait until all SIB bytes were received:
loop_writeEeprom_waitSibRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	lds		tmp0, updiUartDataIndex_ram
	cpi		tmp0, UPDI_SIB_LENGTH
	brlo	loop_writeEeprom_waitSibRx
; updiUartDataBuf_ram[10] now contains the ASCII-encoded NVM version number.
	ldiz	2*szNvmVersion
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram+10
	rcall	uart_putChar
	ldiz	2*szLineBreak
	rcall	uart_putSz
	mov		tmp1, tmp0
; === Enter NVM programming mode (as done by 'pymcuprog') ===
	ldiz	2*szEnteringNVMPK
	rcall	uart_putSz
; Hold target device in reset:
	rcall	updi_applyReset
	rcall	updi_enterNvmProgKey
; Confirm key success by reading ASI Key Status Register:
	ldi		tmp0, UPDI_ASI_KEY_STATUS
	rcall	updi_ldcs
; Read answer from target UPDI:
loop_writeEeprom_waitKeyStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeEeprom_waitKeyStatus
	ldiz	2*szAsiKeyStatus
	rcall	uart_putSz
	lds		tmp0, updiUartDataBuf_ram
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
	bst		tmp0, UPDI_NVMPROG_bp
	brts	loop_writeEeprom_nvmKeySuccess
; NVMPROG key transmitting failed.
	ldiz	2*szErrKey
	rcall	uart_putSz
	rcall	updi_releaseReset
	rjmp	loop_exit
loop_writeEeprom_nvmKeySuccess:
; Release reset and wait for unlock:
	rcall	updi_applyReset
	rcall	wait25ms
	rcall	updi_releaseReset
; == Wait until PROGSTART bit is 1 ==
loop_writeEeprom_progStartLoop:
	ldi		tmp0, UPDI_ASI_SYS_STATUS
	rcall	updi_ldcs
; Read answer from target UPDI:
loop_writeEeprom_waitSysStatus:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	loop_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	loop_writeEeprom_waitSysStatus
	lds		tmp0, updiUartDataBuf_ram
	sbrs	tmp0, UPDI_PROGSTART_bp
	rjmp	loop_writeEeprom_progStartLoop
	ldiz	2*szAsiSysStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Do the NVM operation ==
	rcall	updi_waitForNvm
; EEERWR - EEPROM Erase and Write Enable:
	ldi		tmp0, low(NVMCTRL_CTRLA)
	ldi		tmp1, high(NVMCTRL_CTRLA)
	ldi		tmp2, 0x13
	rcall	updi_sts
	cpi		tmp3, 0
	breq	loop_writeEeprom_addrData
	cpi		tmp3, 1
	breq	loop_timeout
	rjmp	loop_exit
loop_writeEeprom_addrData:
	rcall	wait25ms
; Send EEPROM address and data:
	lds		tmp0, writeAddrL_ram
	lds		tmp1, writeAddrH_ram
	lds		tmp2, writeData_ram
	rcall	updi_sts
	cpi		tmp3, 0
	breq	loop_writeEeprom_noCmd
	cpi		tmp3, 1
	breq	loop_timeout
	rjmp	loop_exit
loop_writeEeprom_noCmd:
/*	rcall	updi_waitForNvm
; Send NOCMD:
	ldi		tmp0, low(NVMCTRL_CTRLA)
	ldi		tmp1, high(NVMCTRL_CTRLA)
	ldi		tmp2, 0x00		; NOCMD.
	rcall	updi_sts
	cpi		tmp3, 0
	breq	loop_writeEeprom_closeNvm
	cpi		tmp3, 1
	breq	loop_timeout
	rjmp	loop_exit
loop_writeEeprom_closeNvm:*/
/*	rcall	updi_waitForNvm
	sbrc	tmp1, 0
	rjmp	loop_timeout*/
	rcall	wait25ms
	ldiz	2*szNvmctrlStatus
	rcall	uart_putSz
	rcall	uart_putHexByte
	ldiz	2*szLineBreak
	rcall	uart_putSz
; == Issue a System Reset to the target device ==
	ldiz	2*szReset
	rcall	uart_putSz
	rcall	updi_applyReset
	rcall	updi_idle
	rcall	wait25ms
	rcall	updi_releaseReset
	rcall	updi_idle
	ldiz	2*szLineBreak
	rcall	uart_putSz
	rjmp	loop_exit

loop_timeout:
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrTimeout
	rcall	uart_putSz
; Send BREAK to re-establish synchronization and disable UPDI:
	rcall	updi_break
loop_exit:
	rcall	updi_disable
	rjmp	loop
	