; File:		updi-tiny3216\updi-link.asm
; Device:	ATtiny3216
; Created:	2024-06-25
; Version:	2024-06-25
; Author:	Johannes Fechner

; == UPDI symbol declarations ==
; === UPDI instructions and arguments ===
.equ UPDI_KEY = 0xE0
.equ UPDI_KEY_64BITS = 0x00
.equ UPDI_KEY_128BITS = 0x01
.equ UPDI_KEY_GET_SIB = 1<<2
.equ UPDI_KEY_SEND_KEY = 0x00

; === UPDI register addresses and bit positions ===
.equ UPDI_ASI_KEY_STATUS = 0x07
.equ UPDI_CHIPERASE_bp = 3
.equ UPDI_NVMPROG_bp = 4
.equ UPDI_ASI_SYS_STATUS = 0x0B
.equ UPDI_LOCKSTATUS_bp = 0
.equ UPDI_PROGSTART_bp = 3 ; also called 'NVMPROG'.
.equ UPDI_ERASEFAIL_bp = 6

; === Other UPDI properties ===
.equ UPDI_SIB_LENGTH = 16 ; System Information Block length in bytes.

; == Subroutines ==
; == Send the NVMPROG key ==
updi_enterNvmProgKey:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_enterNvmProgKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_enterNvmProgKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, tmp0
	ldi		tmp0, 0x20				; Key byte 0.
	st		Z+, tmp0
	ldi		tmp0, 0x67				; Key byte 1.
	st		Z+, tmp0
	ldi		tmp0, 0x6F				; Key byte 2.
	st		Z+, tmp0
	ldi		tmp0, 0x72				; Key byte 3.
	st		Z+, tmp0
	ldi		tmp0, 0x50				; Key byte 4.
	st		Z+, tmp0
	ldi		tmp0, 0x4D				; Key byte 5.
	st		Z+, tmp0
	ldi		tmp0, 0x56				; Key byte 6.
	st		Z+, tmp0
	ldi		tmp0, 0x4E				; Key byte 7.
	st		Z+, tmp0
	ldi		tmp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Send the Chip Erase key ==
updi_enterChipEraseKey:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_enterChipEraseKey_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_enterChipEraseKey_wait
	ldiz	updiUartDataBuf_ram
	ldi		tmp0, UPDI_SYNCH
	st		Z+, tmp0
	ldi		tmp0, 0b111_00_0_00	; Send 64-bit key.
	st		Z+, tmp0
	ldi		tmp0, 0x65				; Key byte 0.
	st		Z+, tmp0
	ldi		tmp0, 0x73				; Key byte 1.
	st		Z+, tmp0
	ldi		tmp0, 0x61				; Key byte 2.
	st		Z+, tmp0
	ldi		tmp0, 0x72				; Key byte 3.
	st		Z+, tmp0
	ldi		tmp0, 0x45				; Key byte 4.
	st		Z+, tmp0
	ldi		tmp0, 0x4D				; Key byte 5.
	st		Z+, tmp0
	ldi		tmp0, 0x56				; Key byte 6.
	st		Z+, tmp0
	ldi		tmp0, 0x4E				; Key byte 7.
	st		Z+, tmp0
	ldi		tmp0, 10	; Count of bytes to be sent.
	sts		updiUartDataCount_ram, tmp0
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Apply reset to the target device ==
updi_applyReset:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_applyReset_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_applyReset_wait
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	rcall	updi_appendBuf
	ldi		tmp0, 0x59					; Normal Reset signature.
	rcall	updi_appendBuf
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret
	
; == Release reset from the target device ==
updi_releaseReset:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_releaseReset_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_releaseReset_wait
; Clear the system reset:
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, 0b110_0_0000 | 0x08	; STCS ASI_RESET_REQ.
	rcall	updi_appendBuf
	ldi		tmp0, 0x00					; Clear reset.
	rcall	updi_appendBuf
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Send LDCS instruction ==
; Parameters:	tmp0, the UPDI register to be loaded.
;				Must be in range 0x00..0x0F.
updi_ldcs:
	push	tmp0
	push	tmp1
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_ldcs_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_ldcs_wait
	mov		tmp1, tmp0
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	andi	tmp1, 0x0F
	ori		tmp1, 0b100_0_0000	; LDCS instruction.
	mov		tmp0, tmp1
	rcall	updi_appendBuf
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp1
	pop		tmp0
	ret

; == Send LDS instruction ==
; Parameters:	YL, address LSB,
;				YH, address MSB.
updi_lds:
	push	tmp0
	push	ZL
	push	ZH
; Wait for any previous transmission or reception to be completed:
updi_lds_wait:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_lds_wait
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, 0b000_0_01_00			; LDS instruction (word address, 1 byte of data).
	rcall	updi_appendBuf
	mov		tmp0, YL					; Address LSB.
	rcall	updi_appendBuf
	mov		tmp0, YH					; Address MSB.
	rcall	updi_appendBuf
	rcall	updi_tx
	pop		ZH
	pop		ZL
	pop		tmp0
	ret

; == Send STS instruction ==
; Parameters:	tmp0, address LSB,
;				tmp1, address MSB,
;				tmp2, data.
; Return:		tmp3, 0: success, 1: timeout, 2: invalid answer.
updi_sts:
	push	tmp0
	push	tmp3
	push	ZL
	push	ZH
; For UART debugging:
	ldiz	2*szSendingSts
	rcall	uart_putSz
; Wait for any previous transmission or reception to be completed:
updi_sts_waitBusy:
	sbrc	updiFlags, UPDI_FLAGS_BUSY_bp
	rjmp	updi_sts_waitBusy
; Send instruction code and address:
	sts		updiUartDataCount_ram, zero
	mov		tmp3, tmp0
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	ldi		tmp0, 0b010_0_01_00		; STS instruction (word address, 1 byte of data).
	rcall	updi_appendBuf
	mov		tmp0, tmp3				; Address LSB.
	rcall	updi_appendBuf
	mov		tmp0, tmp1				; Address MSB.
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for first 'ACK' from Target UPDI:
updi_sts_wait0:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	updi_sts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	updi_sts_wait0
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
    cpi		tmp0, UPDI_ACK
	brne	updi_sts_invAnswer
; Got first ACK. Send data byte:
	rcall	wait1ms
	sts		updiUartDataCount_ram, zero
	ldi		tmp0, UPDI_SYNCH
	rcall	updi_appendBuf
	mov		tmp0, tmp2				; Data byte to be written.
	rcall	updi_appendBuf
	rcall	updi_tx
; Wait for second 'ACK' from Target UPDI:
updi_sts_wait1:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	updi_sts_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	updi_sts_wait1
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
    cpi		tmp0, UPDI_ACK
    brne	updi_sts_invAnswer
; Got second ACK.
	ldi		tmp3, 0
	ldiz	2*szOK
	rcall	uart_putSz
	rjmp	updi_sts_pop
updi_sts_invAnswer:
	ldi		tmp3, 2
	ldiz	2*szLineBreak
	rcall	uart_putSz
	ldiz	2*szErrInvAnswer
	rcall	uart_putSz
	rjmp	updi_sts_pop
updi_sts_timeout:
	ldi		tmp3, 1
updi_sts_pop:
	pop		ZH
	pop		ZL
	pop		tmp3
	pop		tmp0
	ret

; == Wait while NVMCTRL is busy ==
; Return:		tmp0, last value of NVMCTRL_STATUS.
;				tmp1, 0x00: success, 0x01: timeout.
updi_waitForNvm:
	push	tmp2
	push	tmp3
	push	tmp4
	ldiz	2*szWaitNvm
	rcall	uart_putSz
	ldi		YL, low(NVMCTRL_STATUS)
	ldi		YH, high(NVMCTRL_STATUS)
updi_waitForNvm_loop:
	rcall	updi_idle
	rcall	wait1ms
	rcall	updi_lds
; Read answer byte from Target UPDI:
updi_waitForNvm_waitRx:
	sbrc	updiFlags, UPDI_FLAGS_TIMEOUT_bp
	rjmp	updi_waitForNvm_timeout
	sbrs	updiFlags, UPDI_FLAGS_RXC_bp
	rjmp	updi_waitForNvm_waitRx
; updiUartData_ram now contains the received byte.
    lds     tmp0, updiUartDataBuf_ram
	mov		tmp1, tmp0
	andi	tmp1, NVMCTRL_EEBUSY_bm | NVMCTRL_FBUSY_bm
	brne	updi_waitForNvm_loop		; Branch if EEBUSY or FBUSY was set.
	ldi		tmp1, 0
	rjmp	updi_waitForNvm_pop
updi_waitForNvm_timeout:
	ldi		tmp1, 1
updi_waitForNvm_pop:
	pop		tmp4
	pop		tmp3
	pop		tmp2
	ret
